﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using com.ihr.xbrl.om;
using com.ihr.xbrl.om.xdt;
using java.io;
using java.net;
using java.util;
using com.ihr.xbrl.om.instance;
using com.ihr.xbrl.om.formula;
using javax.xml.@namespace;
using com.ihr.xbrl.om.taxonomy;
using com.ihr.xbrl.om.exLinks;

namespace DimensionsTheDTS
{
    class Program
    {
        static void Main(string[] args)
        {
            System.Console.WriteLine("Starting...");
            // the main purpose of this demo is to show how to obtain dimensional information of a fact in an instance document.

            // we will:
            // 1. open an instance document that contains facts that in turn, contains dimensional information.
            // 2. validate the document against the XBRL 2.1 and Dimensions specifications.
            // 3. loop over the instance document content and show fact, value, dimensions and dimensional information.

            // Let's define the properties of the DTSContainer first. By default, the DTSContainer will perform only XBRL 2.1 validation.
            // we can easily configure the processor to perform XBRL 2.1 and Dimensions validation.

            Properties dtsProps = new Properties();
            dtsProps.setProperty(DTSContainer.PROCESSORS_SEQUENCE,
                XBRLCoreProcessor.PROCESSORNAME + "," + XDTProcessor.PROCESSORNAME);

            // Hint: adding more processors to the sequence is possible (Formulas) it is also possible creating custom
            // processors for validation of project specific things that are not part of the XBRL and cannot be part
            // of the Formulas, Dimensions or core XBRL 2.1 and attach them to the sequence.

            // Now lets create the URL of the instance document file
            // change this if the file is in another directory.
            // The selected instance document is using COREP taxonomies 1.3 and sadly, the instance document is pointing to the
            // taxonomies using "relative" locations. This means you will get DTSDiscoveryExceptions if the taxonomy files are
            // not in the same directory as the instance document.
            // well designed instance documents shall refer to the taxonomies at their official URL locations on the Internet.
            // that would allow users to open the instance document whithout having worry about where the taxonomy is.
            // Other XBRL Sw vendors simply do not know about how important this is and does not facilitate the use of XBRL
            // making things more complicated to the end users.
            File fInstance = new File("TaxonomyFiles\\t-ca-2006-07-01.xsd");
            URI uInstance = fInstance.toURI(); // it is easier this way because URIs are more difficult to write when they are pointing to local files.

            // create the DTSContainer using indicated properties
            DTSContainer myDTS = DTSContainer.newEmptyContainer(dtsProps);

            System.Console.Write("Ready; ");
            System.Console.WriteLine("Press any key to start loading the taxonomy...");
            System.Console.ReadKey(true);
            System.Console.WriteLine("Loading... this may take a while...");

            // note this cast is not safe and will fail if the document pointed to by the URL is a linkbase or instance. But this is a demo.
            XBRLTaxonomy myTaxonomy = (XBRLTaxonomy)myDTS.load(uInstance);

            System.Console.WriteLine("Loaded; Press any key to start validating the taxonomy...");
            System.Console.ReadKey(true);
            System.Console.WriteLine("Validating...");

            // Now validate the DTS. This will loop all processor indicated in the PROCESSORS_SEQUENCE property
            // and will collect everything found (Errors, Inconsistencies, Assertions, Formula, ... into the XBRLValidationReport
            XBRLValidationReport report = myDTS.validate();

            // in case of errors. we cannot continue. this call will thow the first error as an XBRLValidationExcecption and will do
            // nothing if the document is clean. Note we don't mind about inconsistencies at this stage. All of them are collected
            // inside the XBRLValidationReport
            report.throwExceptions();

            System.Console.WriteLine("Validated; Press any key to start exploring the instance content...");
            System.Console.ReadKey(true);
            System.Console.WriteLine("Exploring...");

            // obtain the instance of the dimensions processor used during dimensions validation. 
            // (or creates a new one if it were not used) this is the safe way to obtain the instance
            // of the XDTProcessor.
            XDTProcessor xdt = XDTProcessor.create(myDTS);

            // all this is only needed for method 2.
            XDTInfosetConsumer ic = new XDTInfosetConsumer(xdt);

            // you could also do. 
            // XDTProcessor xdt2 = (XDTProcessor)myDTS.getProcessor(XDTProcessor.PROCESSORNAME);
            // and check if the return value is null. It is null if the processor has not been created before.
            // this is the reason the previous method always success.

            // now let's explore the report content. Loop over all existing facts
            Iterator iterDims = xdt.getDimensions();
            while (iterDims.hasNext())
            {
                XDTDimension dim = (XDTDimension)iterDims.next();
                System.Console.WriteLine("Dimension " + printQName(dim.getItem().getQName()));

                // Warning,
                // The same dimension may be used in different hypercubes and on each hypercube
                // the same dimension may have different domain members.
                // 
                // as documented in the API, if the dimension is explicit
                // the getDomain() member will return the list of members associated with the
                // dimension. But, this is NOT the way it should be unless you are absolutely
                // sure the dimension domain is fixed and "the same" all times the dimension
                // is used in hypercubes.
                //
                // For this reason, we are going to show here how the dimension domain should
                // be obtained from the dimension uses in hypercubes in the DTS.

                // first, the loop starts with primary items in the DTS. Primary items 
                // use Hypercubes, that in turn, use dimensions.
                Iterator iterPri = xdt.getPrimaryItems();
                while (iterPri.hasNext())
                {
                    QName priItemQName = (QName)iterPri.next();
                    XDTPrimaryItem priItem = xdt.getPrimaryItem(priItemQName);

                    // a primary item may have hypercubes in multiple Extended links sets (called Base Sets in the API)
                    Iterator iterBase = priItem.getBases();
                    while (iterBase.hasNext())
                    {
                        DTSBase baseSet = (DTSBase)iterBase.next();
                        // in this base set, there may by multiple hypercubes
                        Iterator iterHc = priItem.getHypercubes(baseSet);
                        while (iterHc.hasNext())
                        {
                            XDTHypercubeUse hu = (XDTHypercubeUse)iterHc.next();
                            // now we have the hypercube. let's explore the dimensions it has
                            Iterator iterDim = hu.getDimensions();
                            while (iterDim.hasNext())
                            {
                                XDTDimensionUse du = (XDTDimensionUse)iterDim.next();
                                XDTDimension dimDef = du.getDimension();
                                if (dimDef.equals(dim))
                                {
                                    // we have found a dimension use for a primary item
                                    // in the DTS via the hypercube that is using this
                                    // dimension. The XDTDimensionUse object contains
                                    // the usable domain and everything that belongs
                                    // to this particular dimension under this hypercube

                                    System.Console.WriteLine("Used on " + printQName(priItem.getItem().getQName()) +
                                        " using hypercube " + hu.getName());
                                }
                            }
                        }
                    }
                }
            }

            System.Console.WriteLine("Finished; Press any key to exit...");
            System.Console.ReadKey(true);
        }



        static string printQName(QName q)
        {
            return (q.getPrefix().Length == 0 ? "" : q.getPrefix() + ":") + q.getLocalPart();
        }

    }
}
